#!/usr/bin/perl -s
#
# mcp.pl   Mirror Copy - Jim Bumgardner, 2005
#
# copy one or more files to the most appropriate parallel directory on some other server, using scp
#
# uses .mcp files to determine where files should be copied
#
# example usage:  mcp.pl myfile.xml
#
# an .mcp file should be located either in the current directory, or an ancestor directory
# of the current directory.
#
# this file contains one line for each destination that files should be copied to, showing the
# path to the parallel directory on the other server.
#
# for example, in my local directory /home/jbum/dev/yahoo/properties/music/websites/www/v2/share/htdocs/f8
# there is an .mcp file which reads:
#
# jbum@dev1.music.yahoo.com:/homes/jbum/dev/yahoo/properties/music/websites/www/v2/share/htdocs/f8
# jbum@glossaryyes-df.santamonica.yahoo.com:/home/jbum/dev/yahoo/properties/music/websites/www/v2/share/htdocs/f8
#
# this causes all files in f8/ or below to be copied to those two servers, in a matching directory below f8/


$syntax_error = <<EOT;
mcp [options] file [file...]

options:
    -t     Test mode.  Don't actually copy the files - just show the commands that would be generated.
    -r     Reverse mode.  Copy the file from the remote server to you.
    -n=#   Only execute the Nth command, where # is the command number (first command is 1).

EOT

die "No files specified\n\n" . $syntax_error  if (@ARGV == 0);

$currentDir = `pwd`;

chop $currentDir;

$localRootDir = $currentDir;

# find the closest relevant .mcp file
while (length($localRootDir) && !(-e "$localRootDir/.mcp"))
{
    $localRootDir =~ s~/[^/]+$~~;
}

# if the file exists
my $lineNbr = 0;

if (-e "$localRootDir/.mcp")
{
    # use this as our base local directory
    $currentDir =~ s~^$localRootDir~~;

    # loop on each line
    open (IFILE, "$localRootDir/.mcp") or die ("Can't open $localRootDir/.mcp");
    while (<IFILE>) 
    {
        next if (/^\s*[\#;]/);  # skip comments
        next if (/^\s*$/);      # skip blanks

        # determine where to copy stuff
        $remoteRootDir = $_;
        chop $remoteRootDir;
        
        # copy each file specified on the command line
        foreach $file (@ARGV)
        {
            my ($exdir) = $file =~ m~(.*/)~;
            my $dest = $remoteRootDir . $currentDir;
            my $destD = $dest . ($exdir ne ''? '/' . $exdir : '');
            my $cmd = "scp \'$file\' $destD";
            $cmd = "scp \'$destD/$file\' ." if ($g || $r || $get);
            ++$lineNbr;
            if (!$n || $n == $lineNbr)
            {
                    print $cmd . "\n";
                    system($cmd) if !(defined $t);
            }
        }
    }
}
else {
    die ("Not in a recognizable directory (need at least one .mcp file in current path)\n");
}


